
///////////////////////////////////////////////////////////////////////////////
//
//	level_sprui.cpp
//

#include <stdio.h>
#include "id.h"
#include "project.h"

#ifdef _MSC_VER
#include <uxtheme.h>
#endif

//-----------------------------------------------------------------------------

void LevelSpritesManager::get_sprite_text(char *dst, size_t size, const SpriteAllocData *sprite){
	char s[128];
	get_sprite_name(s, sizeof s, sprite->type);
	sprintf_s(dst, size, "%04X-%04X: %s", sprite->param, sprite->type, s);
	return;
}

void LevelSpritesManager::get_sprite_name(char *dst, size_t size, WORD type){
	char s[16];
	sprintf_s(s, sizeof s, "%04X", type);
	GetPrivateProfileString("SpriteName", s, "?", dst, size, SpriteInfoFileName);
	return;
}

LevelSpritesManager::LevelSpritesManager(LevelManager *_level, ProjectManager *_project)
: level(_level), project(_project), bDragging(false)
{
	hilSpriteImage	= NULL;
	hpnSpriteLine	= NULL;
	hpnSpriteSel	= NULL;
	hbrSpriteRect	= NULL;
	hbrSpriteLine	= NULL;
	hbrSpriteSel	= NULL;
	hftSpriteText	= NULL;
}

LevelSpritesManager::~LevelSpritesManager(){
	if (hpnSpriteLine)	DeleteObject(hpnSpriteLine);
	if (hpnSpriteSel)	DeleteObject(hpnSpriteSel);
	if (hbrSpriteRect)	DeleteObject(hbrSpriteRect);
	if (hbrSpriteLine)	DeleteObject(hbrSpriteLine);
	if (hftSpriteText)	DeleteObject(hftSpriteText);
	if (hilSpriteImage)	ImageList_Destroy(hilSpriteImage);
}

BOOL LevelSpritesManager::create_sys(){
	hilSpriteImage	= ImageList_Create(256, 256, ILC_COLOR24 | ILC_MASK, 128, 0);
	hpnSpriteLine	= CreatePen(PS_SOLID, 1, RGB(255, 0, 0));
	hpnSpriteSel	= CreatePen(PS_SOLID, 3, RGB(255, 128, 128));
	hbrSpriteRect	= CreateSolidBrush(RGB(0, 0, 0));
	hbrSpriteLine	= CreateSolidBrush(RGB(255, 0, 0));
	hbrSpriteSel	= CreateSolidBrush(RGB(128, 0, 0));
	hftSpriteText	= CreateFont(12, 0, 0, 0, FW_REGULAR, FALSE, FALSE, FALSE, SHIFTJIS_CHARSET,
		OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS, PROOF_QUALITY, FIXED_PITCH | FF_MODERN, "lr SVbN");
	iShowSprites	= (ShowSprites)GetPrivateProfileInt("MapView", "ShowSprites", TRUE, ConfigFileName);
	set_menuitem_state();
	return TRUE;
}

BOOL LevelSpritesManager::create(HWND hwTab){
	if (!create_dialog(hwTab)) return FALSE;
	return TRUE;
}

void LevelSpritesManager::set_menuitem_state(){
	MENUITEMINFO mi;
	mi.cbSize	= sizeof mi;
	mi.fMask	= MIIM_STATE;
	mi.fState	= iShowSprites==SPR_HIDE ? MFS_CHECKED : MFS_UNCHECKED;
	SetMenuItemInfo(GetMenu(project->get_hWnd()), IDM_VIEW_SPR_HIDE, FALSE, &mi);
	mi.fState	= iShowSprites==SPR_NORMAL ? MFS_CHECKED : MFS_UNCHECKED;
	SetMenuItemInfo(GetMenu(project->get_hWnd()), IDM_VIEW_SPR_NORMAL, FALSE, &mi);
	mi.fState	= iShowSprites==SPR_SIMPLE ? MFS_CHECKED : MFS_UNCHECKED;
	SetMenuItemInfo(GetMenu(project->get_hWnd()), IDM_VIEW_SPR_SIMPLE, FALSE, &mi);
	return;
}

void LevelSpritesManager::save_data(){
	char s[16];
	sprintf_s(s, sizeof s, "%d", (int)iShowSprites);
	WritePrivateProfileString("MapView", "ShowSprites", s, ConfigFileName);
	return;
}

void LevelSpritesManager::clear(){
	LPCSTR empty = "";
	data.clear();
	SendMessage(GetDlgItem(hWnd, IDC_LIST0), LB_RESETCONTENT, 0L, 0L);
	unable_edit_change = true;
	SetDlgItemText(hWnd, IDC_EDIT0, empty);
	SetDlgItemText(hWnd, IDC_EDIT1, empty);
	SetDlgItemText(hWnd, IDC_EDIT2, empty);
	SetDlgItemText(hWnd, IDC_EDIT3, empty);
	SetDlgItemText(hWnd, IDC_EDIT4, empty);
	SetDlgItemText(hWnd, IDC_EDIT5, empty);
	unable_edit_change = false;
	return;
}

void LevelSpritesManager::load(int _level_num){
	WORD left, right, sdr;
	char s[128], s2[128];
	level_num = _level_num;
	clear();
	// XvCgzu̓ǂݍ
	int level_sprites_count = GetPrivateProfileInt("Sprites", "LevelCount", 224, RomInfoFileName);
	SADRS adrs_sprites = project->read_sadrs(0x3E, 2*level_num);
	if (data.load_data(project, 0x3E, adrs_sprites) == -1){
		MessageBox(project->get_hWnd(), "XvCgf[^̓ǂݍ݂Ɏs܂B", "G[",
			MB_OK | MB_ICONEXCLAMATION);
	}else{
		right = adrs_sprites + data.get_size();
		left = GetPrivateProfileInt("Sprites", "AvailableEnd", 0x9500, RomInfoFileName);
		for (int i=0; i<level_sprites_count; i++){
			sdr = project->read_sadrs(0x3E, 2*i);
			if (sdr >= right && sdr < left) left = sdr;
		}
		max_size = left - adrs_sprites;
	}
	LevelSpritesData::iterator i, i_end = data.end();
	// C[WXg̍쐬
	sprite_type_map.clear();
	sprite_image_map.clear();
	ImageList_RemoveAll(hilSpriteImage);
	for (i=data.begin(); i!=i_end; i++){
		prepare_sprite_image((*i).type);
	}
	// Xg̍XV
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	SendMessage(hwList, LB_RESETCONTENT, 0, 0);
	int n;
	for (i=data.begin(), n=0; i!=i_end; i++, n++){
		get_sprite_text(s, sizeof s, &(*i));
		sprintf_s(s2, sizeof s2, "#%d: %s", n, s);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)n, (LPARAM)s2);
	}
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	isel = -1;
	return;
}

void LevelSpritesManager::prepare_sprite_image(WORD sprite_type){
	static BITMAPINFOHEADER bih = { sizeof(BITMAPINFOHEADER), 256, 256, 1, 24, BI_RGB, 256*256*3, 0, 0, 0, 0 };
	static BYTE image_buf[256][256][3];
	SpriteTypeMap::iterator itr_type = sprite_type_map.find((int)sprite_type);
	SpriteTypeMap::iterator itr_type_end = sprite_type_map.end();
	if (itr_type != sprite_type_map.end()) return;	// ̎
	WORD dynamic_num, palette_num;
	bool h_flip, v_flip;
	if (!project->sprites.get_sprite_graphics(sprite_type, &dynamic_num, &palette_num, &h_flip, &v_flip)) return;	// 摜擾
	sprite_type_map.insert(SpriteTypeElement((int)sprite_type, MakeSpriteLONG(dynamic_num, palette_num, h_flip, v_flip)));
	SpriteImageMap::iterator itr = sprite_image_map.find(MakeSpriteLONG(dynamic_num, palette_num, h_flip, v_flip));
	if (itr != sprite_image_map.end()) return;			// ̉摜
	WORD *palette = project->image.get_palette(palette_num);
	WORD static_num = 0;
	for (DynamicImageIterator dyn_itr = project->image.get_dynamic(dynamic_num); dyn_itr; dyn_itr++){
		if ((dyn_itr.get_command() & 0x80) == 0){
			static_num = dyn_itr.get_value();			// ŏ̐Î~̗p
			if (static_num != 0) break;					// A0ꍇ͂ƒT
		}
	}
	if (static_num == 0) return;					// Î~悪Ȃ
	LADRS static_adrs = project->image.get_static_image(static_num);
	if (static_adrs.bank >= 0x40) return;				// sANZXh~
	SpriteImage si;
	si.rtClip.left   = 0xFFFF;
	si.rtClip.top    = 0xFFFF;
	si.rtClip.right  = 0;
	si.rtClip.bottom = 0;
	StaticImageHeader8 header;
	header.load(project, static_adrs);
	int chip_index=0;
	memset(image_buf, 0xFF, 256*256*3);
	WORD *pos_array  = (WORD*)project->get_pointer(static_adrs.bank, static_adrs.adrs + 8);
	BYTE *pattern_base = (BYTE*)project->get_pointer(static_adrs.bank, static_adrs.adrs);
	for (int i=0; i<header.big_chip_count; i++, chip_index++){
		int pos_x = LOBYTE(pos_array[chip_index]);
		int pos_y = HIBYTE(pos_array[chip_index]);
		if (pos_x < si.rtClip.left) si.rtClip.left = pos_x;
		if (pos_y < si.rtClip.top)  si.rtClip.top  = pos_y;
		if (pos_x+16 > si.rtClip.right)  si.rtClip.right = pos_x+16;
		if (pos_y+16 > si.rtClip.bottom) si.rtClip.bottom = pos_y+16;
		draw_mini_chip(image_buf, pos_x, pos_y, pattern_base + header.get_pattern_big0(i), palette, h_flip, v_flip);
		draw_mini_chip(image_buf, pos_x + 8, pos_y, pattern_base + header.get_pattern_big1(i), palette, h_flip, v_flip);
		draw_mini_chip(image_buf, pos_x, pos_y + 8, pattern_base + header.get_pattern_big2(i), palette, h_flip, v_flip);
		draw_mini_chip(image_buf, pos_x + 8, pos_y + 8, pattern_base + header.get_pattern_big3(i), palette, h_flip, v_flip);
	}
	for (int i=0; i<header.small_chip_count; i++, chip_index++){
		int pos_x = LOBYTE(pos_array[chip_index]);
		int pos_y = HIBYTE(pos_array[chip_index]);
		if (pos_x < si.rtClip.left) si.rtClip.left = pos_x;
		if (pos_y < si.rtClip.top)  si.rtClip.top  = pos_y;
		if (pos_x+8 > si.rtClip.right)  si.rtClip.right = pos_x+8;
		if (pos_y+8 > si.rtClip.bottom) si.rtClip.bottom = pos_y+8;
		draw_mini_chip(image_buf, pos_x, pos_y, pattern_base + header.get_pattern_small(i), palette, h_flip, v_flip);
	}
	for (int i=0; i<header.third_chip_count; i++, chip_index++){
		int pos_x = LOBYTE(pos_array[chip_index]);
		int pos_y = HIBYTE(pos_array[chip_index]);
		if (pos_x < si.rtClip.left) si.rtClip.left = pos_x;
		if (pos_y < si.rtClip.top)  si.rtClip.top  = pos_y;
		if (pos_x+8 > si.rtClip.right)  si.rtClip.right = pos_x+8;
		if (pos_y+8 > si.rtClip.bottom) si.rtClip.bottom = pos_y+8;
		draw_mini_chip(image_buf, pos_x, pos_y, pattern_base + header.get_pattern_third(i), palette, h_flip, v_flip);
	}
	if (si.rtClip.left == 0xFFFF || si.rtClip.top == 0xFFFF || si.rtClip.right == 0 || si.rtClip.bottom == 0){
		// p^[\s
		return;
	}
	if (h_flip){
		int temp = si.rtClip.right;
		si.rtClip.right = 256 - si.rtClip.left;
		si.rtClip.left  = 256 - temp;
	}
	if (v_flip){
		int temp = si.rtClip.bottom;
		si.rtClip.bottom = 256 - si.rtClip.top;
		si.rtClip.top    = 256 - temp;
	}
	HBITMAP hBitmap = CreateDIBitmap(GetDC(level->get_hWnd()), &bih, CBM_INIT, image_buf, (LPBITMAPINFO)&bih, DIB_RGB_COLORS);
	si.index = ImageList_AddMasked(hilSpriteImage, hBitmap, RGB(255, 255, 255));
	sprite_image_map.insert(SpriteImageElement(MakeSpriteLONG(dynamic_num, palette_num, h_flip, v_flip), si));
	DeleteObject(hBitmap);
}

void LevelSpritesManager::draw_mini_chip(BYTE buf[256][256][3], int pos_x, int pos_y,
	const BYTE *data, const WORD *palette, bool h_flip, bool v_flip)
{
	for (int iy=0; iy<8; iy++){
		for (int ix=0; ix<8; ix++){
			BYTE bit0 = data[iy*2];
			BYTE bit1 = data[iy*2 + 1];
			BYTE bit2 = data[iy*2 + 16];
			BYTE bit3 = data[iy*2 + 17];
			int c = ((bit0 >> (7-ix)) & 1) | (((bit1 >> (7-ix)) & 1) << 1)
				| (((bit2 >> (7-ix)) & 1) << 2) | (((bit3 >> (7-ix)) & 1) << 3);
			if (c){
				buf[v_flip? pos_y+iy : 255-(pos_y+iy)][h_flip? 255-(pos_x+ix) : pos_x+ix][2] = (palette[c-1] & 0x001F) << 3;
				buf[v_flip? pos_y+iy : 255-(pos_y+iy)][h_flip? 255-(pos_x+ix) : pos_x+ix][1] = (palette[c-1] & 0x03E0) >> 2;
				buf[v_flip? pos_y+iy : 255-(pos_y+iy)][h_flip? 255-(pos_x+ix) : pos_x+ix][0] = (palette[c-1] & 0x7C00) >> 7;
			}else{
				// pbgɕ\Ȃ(255, 255, 255)𗘗p
				buf[v_flip? pos_y+iy : 255-(pos_y+iy)][h_flip? 255-(pos_x+ix) : pos_x+ix][0] = 255;
				buf[v_flip? pos_y+iy : 255-(pos_y+iy)][h_flip? 255-(pos_x+ix) : pos_x+ix][1] = 255;
				buf[v_flip? pos_y+iy : 255-(pos_y+iy)][h_flip? 255-(pos_x+ix) : pos_x+ix][2] = 255;
			}
		}
	}
}

void LevelSpritesManager::set_state(bool bEnabled){
	EnableWindow(GetDlgItem(hWnd, IDC_LIST0), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT0), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT1), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT2), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT3), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT4), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_EDIT5), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON0), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON1), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON2), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON3), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON4), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON5), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_BUTTON6), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_STATIC0), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_STATIC1), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_STATIC2), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_STATIC3), bEnabled);
	EnableWindow(GetDlgItem(hWnd, IDC_STATIC4), bEnabled);
	return;
}

BOOL LevelSpritesManager::Proc(HWND hDlg, UINT msg, WPARAM wParam, LPARAM lParam){
	switch (msg){
		case WM_INITDIALOG:
			{
#ifdef _MSC_VER
				HMODULE hMod;
				hMod = LoadLibrary("uxtheme.dll");
				if (hMod)
				{
					typedef HRESULT (WINAPI *EnableThemeDialogTexturePROC)(HWND, DWORD);
					EnableThemeDialogTexturePROC pEnableThemeDialogTexture;
					pEnableThemeDialogTexture = (EnableThemeDialogTexturePROC)GetProcAddress(hMod, "EnableThemeDialogTexture");
					if (pEnableThemeDialogTexture)
						pEnableThemeDialogTexture(hWnd, ETDT_ENABLETAB);
					FreeLibrary(hMod);
				}
#endif
				SetWindowLong(GetDlgItem(hDlg, IDC_EDIT0), GWL_EXSTYLE, 0);
				//SetWindowLong(GetDlgItem(hDlg, IDC_EDIT5), GWL_EXSTYLE, 0);
				SendDlgItemMessage(hDlg, IDC_EDIT1, EM_SETLIMITTEXT, 4, 0);
				SendDlgItemMessage(hDlg, IDC_EDIT4, EM_SETLIMITTEXT, 4, 0);
				unable_edit_change = false;
			}
			return TRUE;
		case WM_COMMAND:
			switch (LOWORD(wParam)){
				case IDC_LIST0:
					if (HIWORD(wParam) == LBN_SELCHANGE){
						int sel = SendMessage(GetDlgItem(hDlg, IDC_LIST0), LB_GETCURSEL, 0, 0);
						sel_changed(sel);
						if (iShowSprites != SPR_HIDE){
							isel = sel;
							level->excluse_pages(LEVEL_SPRITES);
							level->redraw_default();
						}
						return TRUE;
					}
					if (HIWORD(wParam) == LBN_DBLCLK){
						jump(SendDlgItemMessage(hDlg, IDC_LIST0, LB_GETCURSEL, 0, 0));
						return TRUE;
					}
					break;
				case IDC_EDIT1:
					if (HIWORD(wParam) == EN_CHANGE && !unable_edit_change){
						SetDlgItemText(hWnd, IDC_STATIC1, "p[^(&P)*:");
						return TRUE;
					}
					break;
				case IDC_EDIT2:
					if (HIWORD(wParam) == EN_CHANGE && !unable_edit_change){
						SetDlgItemText(hWnd, IDC_STATIC2, "XW(&X)*:");
						return TRUE;
					}
					break;
				case IDC_EDIT3:
					if (HIWORD(wParam) == EN_CHANGE && !unable_edit_change){
						SetDlgItemText(hWnd, IDC_STATIC3, "YW(&Y)*:");
						return TRUE;
					}
					break;
				case IDC_EDIT4:
					if (HIWORD(wParam) == EN_CHANGE){
						if (!unable_edit_change){
							SetDlgItemText(hWnd, IDC_STATIC4, "(&T)*:");
						}
						type_changed();
						return TRUE;
					}
					break;
				case IDC_BUTTON0:	// ...
					select_sprite();
					return TRUE;
				case IDC_BUTTON1:	// Kp
					spr_apply(SendDlgItemMessage(hDlg, IDC_LIST0, LB_GETCURSEL, 0, 0));
					return TRUE;
				case IDC_BUTTON2:	// ǉ
					spr_add(SendDlgItemMessage(hDlg, IDC_LIST0, LB_GETCURSEL, 0, 0));
					return TRUE;
				case IDC_BUTTON3:	// 폜
					spr_delete(SendDlgItemMessage(hDlg, IDC_LIST0, LB_GETCURSEL, 0, 0));
					return TRUE;
				case IDC_BUTTON4:	// 
					spr_up(SendDlgItemMessage(hDlg, IDC_LIST0, LB_GETCURSEL, 0, 0));
					return TRUE;
				case IDC_BUTTON5:	// 
					spr_down(SendDlgItemMessage(hDlg, IDC_LIST0, LB_GETCURSEL, 0, 0));
					return TRUE;
				case IDC_BUTTON6:	// ɂ
					spr_sort(SendDlgItemMessage(hDlg, IDC_LIST0, LB_GETCURSEL, 0, 0));
					return TRUE;
			}
			break;
			/*
		case WM_CTLCOLORDLG:
			return (BOOL)GetStockObject(NULL_BRUSH);
		case WM_CTLCOLORBTN:
		case WM_CTLCOLOREDIT:
		//case WM_CTLCOLORLISTBOX:
		//case WM_CTLCOLORMSGBOX:
		//case WM_CTLCOLORSCROLLBAR:
		case WM_CTLCOLORSTATIC:
			//break;
			return (BOOL)GetStockObject(NULL_BRUSH);//*/
	}
	return FALSE;
}

void LevelSpritesManager::select_sprite(){
	char s[256];
	int val;
	GetDlgItemText(hWnd, IDC_EDIT4, s, sizeof s);
	if (sscanf_s(s, "%x", &val) < 1) val = -1;
	val = project->sprites.sprite_select_dialog(project->get_hWnd(), val);
	if (val == -1) return;
	sprintf_s(s, sizeof s, "%04X", val);
	SetDlgItemText(hWnd, IDC_EDIT4, s);
	return;
}

void LevelSpritesManager::type_changed(){
	char s[256];
	int val;
	GetDlgItemText(hWnd, IDC_EDIT4, s, sizeof s);
	if (sscanf_s(s, "%x", &val) < 1) val = 0;
	get_sprite_name(s, sizeof s, (WORD)val);
	SetDlgItemText(hWnd, IDC_EDIT5, s);
	return;
}

void LevelSpritesManager::spr_apply(int sel){
	char s[256], s2[256];
	LevelSpritesData::iterator i = get_iterator(sel);
	if (i == data.end()) return;
	SpriteAllocData data_before = *i;
	input_sprite_data(&(*i));
	// ROMf[^ɔf
	SADRS adrs_sprite = project->read_sadrs(0x3E, 2*level_num) + sel*SpriteAllocData::size;
	(*i).store_data(project, 0x3E, adrs_sprite);
	// XvCg摜p
	prepare_sprite_image((*i).type);
	// Xgɔf
	get_sprite_text(s2, sizeof s2, &(*i));
	sprintf_s(s, sizeof s, "#%d: %s", sel, s2);
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	SendMessage(hwList, LB_DELETESTRING, (WPARAM)sel, 0);
	SendMessage(hwList, LB_INSERTSTRING, (WPARAM)sel, (LPARAM)s);
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)sel, 0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	level->redraw_default();
	// Undoobt@֒ǉ
	LevelSpriteApply_UndoMethod *undo = new LevelSpriteApply_UndoMethod(project, level, this,
		level_num, sel, &data_before, &*i);
	project->undo_buffer.add_method(undo);
	return;
}

void LevelSpritesManager::spr_add(int sel){
	char s[256], s2[128];
	int n;
	if (data.get_size() + SpriteAllocData::size > max_size){
		sprintf_s(s, sizeof s, "ȏXvCgǉł܂B"
			"\r\n: %doCg, ő: %doCg", data.get_size(), max_size);
		MessageBox(project->get_hWnd(), s, "G[", MB_OK | MB_ICONEXCLAMATION);
		return;
	}
	LevelSpritesData::iterator i_sel = get_iterator(sel), i, i_end = data.end();
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	for (i=i_sel; i!=i_end; i++){
		SendMessage(hwList, LB_DELETESTRING, (WPARAM)sel, 0);
	}
	// ݓ͂Ălf[^쐬
	SpriteAllocData sprite;
	input_sprite_data(&sprite);
	i_sel = data.insert(i_sel, sprite);
	if (sel == -1){
		sel = data.get_sprite_count()-1;
		get_sprite_text(s2, sizeof s2, &sprite);
		sprintf_s(s, sizeof s, "#%d: %s", sel, s2);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)sel, (LPARAM)s);
	}else{
		for (i=i_sel, n=sel; i!=i_end; i++, n++){
			get_sprite_text(s2, sizeof s2, &(*i));
			sprintf_s(s, sizeof s, "#%d: %s", n, s2);
			SendMessage(hwList, LB_INSERTSTRING, (WPARAM)n, (LPARAM)s);
		}
	}
	if (iShowSprites != SPR_HIDE){
		isel = sel;
		level->excluse_pages(LEVEL_SPRITES);
	}
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)sel, 0);
	output_sprite_data(&*i_sel, sel);
	// ROMɔf
	SADRS adrs_sprites = project->read_sadrs(0x3E, 2*level_num);
	data.store_data(project, 0x3E, adrs_sprites, max_size);
	level->redraw_default();
	// XvCg摜p
	prepare_sprite_image((*i_sel).type);
	// Undoobt@֒ǉ
	LevelSpriteAdd_UndoMethod *undo = new LevelSpriteAdd_UndoMethod(project, level, this,
		level_num, sel, &sprite);
	project->undo_buffer.add_method(undo);
	return;
}

void LevelSpritesManager::spr_delete(int sel){
	char s[128], s2[128];
	int n;
	LevelSpritesData::iterator i_sel = get_iterator(sel), i, i_end = data.end();
	if (i_sel == i_end) return;
	SpriteAllocData sprite = *i_sel;
	// Undoobt@֒ǉ
	LevelSpriteDelete_UndoMethod *undo = new LevelSpriteDelete_UndoMethod(project, level, this,
		level_num, sel, &sprite);
	project->undo_buffer.add_method(undo);
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	for (i=i_sel; i!=i_end; i++){
		SendMessage(hwList, LB_DELETESTRING, (WPARAM)sel, 0);
	}
	if ((i_sel = data.erase(i_sel)) == i_end){
		sel--;
		if (sel != -1){
			output_sprite_data(&(data.back()), sel);
		}
		if (isel != -1) isel = sel;
	}else{
		for (i=i_sel, n=sel; i!=i_end; i++, n++){
			get_sprite_text(s, sizeof s, &(*i));
			sprintf_s(s2, sizeof s2, "#%d: %s", n, s);
			SendMessage(hwList, LB_INSERTSTRING, (WPARAM)n, (LPARAM)s2);
		}
		output_sprite_data(&*i_sel, sel);
	}
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	if (sel != -1){
		SendMessage(hwList, LB_SETCURSEL, (WPARAM)sel, 0);
	}
	// ROMɔf
	SADRS adrs_sprites = project->read_sadrs(0x3E, 2*level_num);
	data.store_data(project, 0x3E, adrs_sprites, max_size);
	level->redraw_default();
	return;
}

void LevelSpritesManager::spr_up(int sel){
	char s[128], s2[128];
	if (sel <= 0) return;
	LevelSpritesData::iterator i_sel = get_iterator(sel), i, i_end = data.end();
	if (i_sel == i_end) return;
	// Undoobt@֒ǉ
	LevelSpriteUp_UndoMethod *undo = new LevelSpriteUp_UndoMethod(project, level, this, level_num, sel);
	project->undo_buffer.add_method(undo);
	SpriteAllocData sprite_temp = *i_sel;
	i = i_sel--;
	*i = *i_sel;
	*i_sel = sprite_temp;
	sel--;
	if (isel != -1) isel--;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	SendMessage(hwList, LB_DELETESTRING, (WPARAM)sel, 0);
	SendMessage(hwList, LB_DELETESTRING, (WPARAM)sel, 0);
	get_sprite_text(s, sizeof s, &(*i_sel));
	sprintf_s(s2, sizeof s2, "#%d: %s", sel, s);
	SendMessage(hwList, LB_INSERTSTRING, (WPARAM)sel, (LPARAM)s2);
	get_sprite_text(s, sizeof s, &(*i));
	sprintf_s(s2, sizeof s2, "#%d: %s", sel+1, s);
	SendMessage(hwList, LB_INSERTSTRING, (WPARAM)(sel+1), (LPARAM)s2);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)sel, 0);
	// ROMɔf
	SADRS adrs_sprite = project->read_sadrs(0x3E, 2*level_num) + sel*SpriteAllocData::size;
	(*i_sel).store_data(project, 0x3E, adrs_sprite);
	(*i).store_data(project, 0x3E, adrs_sprite + SpriteAllocData::size);
	sprintf_s(s, sizeof s, "%d", sel);
	SetDlgItemText(hWnd, IDC_EDIT0, s);
	level->redraw_default();
	return;
}

void LevelSpritesManager::spr_down(int sel){
	char s[128], s2[128];
	if (sel < 0 || sel >= (int)(data.get_sprite_count())-1) return;
	LevelSpritesData::iterator i_sel = get_iterator(sel), i, i_end = data.end();
	if (i_sel == i_end) return;
	// Undoobt@֒ǉ
	LevelSpriteDown_UndoMethod *undo = new LevelSpriteDown_UndoMethod(project, level, this, level_num, sel);
	project->undo_buffer.add_method(undo);
	SpriteAllocData sprite_temp = *i_sel;
	i = i_sel++;
	*i = *i_sel;
	*i_sel = sprite_temp;
	sel++;
	if (isel != -1) isel++;
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	SendMessage(hwList, LB_DELETESTRING, (WPARAM)sel, 0);
	SendMessage(hwList, LB_DELETESTRING, (WPARAM)(sel-1), 0);
	get_sprite_text(s, sizeof s, &(*i));
	sprintf_s(s2, sizeof s2, "#%d: %s", sel-1, s);
	SendMessage(hwList, LB_INSERTSTRING, (WPARAM)(sel-1), (LPARAM)s2);
	get_sprite_text(s, sizeof s, &(*i_sel));
	sprintf_s(s2, sizeof s2, "#%d: %s", sel, s);
	SendMessage(hwList, LB_INSERTSTRING, (WPARAM)sel, (LPARAM)s2);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)sel, 0);
	// ROMɔf
	SADRS adrs_sprite = project->read_sadrs(0x3E, 2*level_num) + sel*SpriteAllocData::size;
	(*i_sel).store_data(project, 0x3E, adrs_sprite);
	(*i).store_data(project, 0x3E, adrs_sprite - SpriteAllocData::size);
	sprintf_s(s, sizeof s, "%d", sel);
	SetDlgItemText(hWnd, IDC_EDIT0, s);
	level->redraw_default();
	return;
}

void LevelSpritesManager::spr_sort(int sel){
	int n, m;
	char s[256], s2[256];
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	LevelSpritesData data_before(data);
	int sel_before = sel;
	SpriteAllocData sprite_temp;
	LevelSpritesData::iterator i1, i2, i_end = data.end();
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	for (i1=data.begin(), n=0; i1!=i_end; i1++, n++){
		for (i2=i1++, m=n+1; i1!=i_end; i1++, m++){
			if (((*i2).pdx == (*i1).pdx) ? ((*i2).pdy > (*i1).pdy) : ((*i2).pdx > (*i1).pdx)){
				sprite_temp = *i2;
				*i2 = *i1;
				*i1 = sprite_temp;
				if (n == sel){
					sel = m;
				}else if (m == sel){
					sel = n;
				}
			}
		}
		i1 = i2;
		SendMessage(hwList, LB_DELETESTRING, (WPARAM)n, 0);
		get_sprite_text(s, sizeof s, &(*i1));
		sprintf_s(s2, sizeof s2, "#%d: %s", n, s);
		SendMessage(hwList, LB_INSERTSTRING, (WPARAM)n, (LPARAM)s2);
	}
	if (isel != -1) isel = sel;
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)sel, 0);
	// ROMɔf
	SADRS adrs_sprites = project->read_sadrs(0x3E, 2*level_num);
	data.store_data(project, 0x3E, adrs_sprites, max_size);
	level->redraw_default();
	if (sel != -1){
		sprintf_s(s, sizeof s, "%d", sel);
		SetDlgItemText(hWnd, IDC_EDIT0, s);
	}
	// Undoobt@֒ǉ
	LevelSpriteSort_UndoMethod *undo = new LevelSpriteSort_UndoMethod(project, level, this, level_num,
		sel_before, sel, data_before, data);
	project->undo_buffer.add_method(undo);
	return;
}

void LevelSpritesManager::jump(int sel){
	LevelSpritesData::iterator i = get_iterator(sel);
	if (i == data.end()) return;
	level->set_scroll_pos((*i).pdx, (*i).pdy);
	return;
}

void LevelSpritesManager::focus_on_sprite(int sel){
	if (iShowSprites != SPR_HIDE){
		isel = sel;
		LevelSpritesData::iterator i = get_iterator(sel);
		if (i == data.end()) return;
		if (!level->is_point_in_screen((*i).pdx, (*i).pdy))
			level->set_scroll_pos((*i).pdx, (*i).pdy);
	}
	level->redraw_default();
	return;
}

void LevelSpritesManager::input_sprite_data(SpriteAllocData *sprite){
	char s[256];
	int int_val;
	GetDlgItemText(hWnd, IDC_EDIT1, s, sizeof s);	// p[^
	if (sscanf_s(s, "%x", &int_val) < 1) int_val = 0;
	(*sprite).param = int_val;
	GetDlgItemText(hWnd, IDC_EDIT2, s, sizeof s);	// X
	if (sscanf_s(s, "%d", &int_val) < 1) int_val = 0;
	(*sprite).pdx = int_val;
	GetDlgItemText(hWnd, IDC_EDIT3, s, sizeof s);	// Y
	if (sscanf_s(s, "%d", &int_val) < 1) int_val = 0;
	(*sprite).pdy = int_val;
	GetDlgItemText(hWnd, IDC_EDIT4, s, sizeof s);	// 
	if (sscanf_s(s, "%x", &int_val) < 1) int_val = 0;
	(*sprite).type = int_val;
	// ږ*
	SetDlgItemText(hWnd, IDC_STATIC1, "p[^(&P):");
	SetDlgItemText(hWnd, IDC_STATIC2, "XW(&X):");
	SetDlgItemText(hWnd, IDC_STATIC3, "YW(&Y):");
	SetDlgItemText(hWnd, IDC_STATIC4, "(&T):");
	return;
}

void LevelSpritesManager::output_sprite_data(const SpriteAllocData *sprite, int iIndex){
	char s[64], s2[256];
	unable_edit_change = true;
	sprintf_s(s, sizeof s, "%d", iIndex);
	SetDlgItemText(hWnd, IDC_EDIT0, s);
	sprintf_s(s, sizeof s, "%04X", (*sprite).param);
	SetDlgItemText(hWnd, IDC_EDIT1, s);
	sprintf_s(s, sizeof s, "%d", (*sprite).pdx);
	SetDlgItemText(hWnd, IDC_EDIT2, s);
	sprintf_s(s, sizeof s, "%d", (*sprite).pdy);
	SetDlgItemText(hWnd, IDC_EDIT3, s);
	sprintf_s(s, sizeof s, "%04X", (*sprite).type);
	SetDlgItemText(hWnd, IDC_EDIT4, s);
	get_sprite_name(s2, sizeof s2, (*sprite).type);
	SetDlgItemText(hWnd, IDC_EDIT5, s2);
	unable_edit_change = false;
	// ږ*
	SetDlgItemText(hWnd, IDC_STATIC1, "p[^(&P):");
	SetDlgItemText(hWnd, IDC_STATIC2, "XW(&X):");
	SetDlgItemText(hWnd, IDC_STATIC3, "YW(&Y):");
	SetDlgItemText(hWnd, IDC_STATIC4, "(&T):");
	return;
}

void LevelSpritesManager::sel_changed(int sel){
	if (sel == -1) return;
	LevelSpritesData::iterator i = get_iterator(sel);
	if (i == data.end()) return;
	output_sprite_data(&(*i), sel);
	return;
}

LevelSpritesData::iterator LevelSpritesManager::get_iterator(int iIndex){
	LevelSpritesData::iterator i, i_end = data.end();
	if (iIndex == -1) return i_end;
	int n;
	for (i=data.begin(), n=0; i!=i_end; i++, n++){
		if (n == iIndex) return i;
	}
	return i_end;
}

bool LevelSpritesManager::mouseover(int pdx, int pdy){
	if (iShowSprites == SPR_HIDE) return false;
	char s[256];
	int x, y, n;
	LevelSpritesData::reverse_iterator i, i_end = data.rend();
	for (i=data.rbegin(), n=data.get_sprite_count()-1; i!=i_end; i++, n--){
		x = (*i).pdx;
		y = (*i).pdy;
		bool bMouseOver = false;
		bool bImageValid = false;
		if (iShowSprites == SPR_NORMAL){
			SpriteTypeMap::iterator itr_type = sprite_type_map.find((int)(*i).type);
			if (itr_type != sprite_type_map.end()){
				SpriteImageMap::iterator itr_image = sprite_image_map.find((*itr_type).second);
				if (itr_image != sprite_image_map.end()){
					bImageValid = true;
					bMouseOver = (pdx >= x-128 + (*itr_image).second.rtClip.left - 8)
						&& (pdx < x-128 + (*itr_image).second.rtClip.right + 8)
						&& (pdy >= y-128 + (*itr_image).second.rtClip.top - 8 - 18)
						&& (pdy < y-128 + (*itr_image).second.rtClip.top - 7);
					bMouseOver |= (pdx >= x-128 + (*itr_image).second.rtClip.left - 8)
						&& (pdx < x-128 + (*itr_image).second.rtClip.left - 8 + 5)
						&& (pdy >= y-128 + (*itr_image).second.rtClip.top - 8) && (pdy < y-128 + (*itr_image).second.rtClip.bottom + 8);
					bMouseOver |= (pdx < x-128 + (*itr_image).second.rtClip.right + 8)
						&& (pdx >= x-128 + (*itr_image).second.rtClip.right + 8 - 5)
						&& (pdy >= y-128 + (*itr_image).second.rtClip.top - 8) && (pdy < y-128 + (*itr_image).second.rtClip.bottom + 8);
					bMouseOver |= (pdy >= y-128 + (*itr_image).second.rtClip.top - 8)
						&& (pdy < y-128 + (*itr_image).second.rtClip.top - 8 + 5)
						&& (pdx >= x-128 + (*itr_image).second.rtClip.left - 8) && (pdx < x-128 + (*itr_image).second.rtClip.right + 8);
					bMouseOver |= (pdy < y-128 + (*itr_image).second.rtClip.bottom + 8)
						&& (pdy >= y-128 + (*itr_image).second.rtClip.bottom + 8 - 5)
						&& (pdx >= x-128 + (*itr_image).second.rtClip.left - 8) && (pdx < x-128 + (*itr_image).second.rtClip.right + 8);
				}
			}
		}
		if (!bImageValid){
			bMouseOver = pdx >= x - SPRITE_WIDTH/2 && pdx < x + SPRITE_WIDTH/2 && pdy >= y-SPRITE_HEIGHT && pdy < y;
		}
		if (bMouseOver){
			SetCursor(LoadCursor(NULL, IDC_SIZEALL));
			sprintf_s(s, sizeof s, "XvCg#%d: (%d, %d)", n, (*i).pdx, (*i).pdy);
			SendMessage(project->get_hwStatus(), SB_SETTEXT, (WPARAM)STATUS_INDEX_POINTING_OBJECT, (LPARAM)s);
			get_sprite_text(s, sizeof s, &(*i));
			SendMessage(project->get_hwStatus(), SB_SETTEXT, (WPARAM)STATUS_INDEX_POINTING_OBJECT_EX, (LPARAM)s);
			return true;
		}
	}
	return false;
}

bool LevelSpritesManager::accepts_click(int pdx, int pdy){
	if (iShowSprites == SPR_HIDE) return false;
	int x, y, n;
	LevelSpritesData::reverse_iterator i, i_end = data.rend();
	for (i=data.rbegin(), n=data.get_sprite_count()-1; i!=i_end; i++, n--){
		x = (*i).pdx;
		y = (*i).pdy;
		bool bMouseOver = false;
		bool bImageValid = false;
		if (iShowSprites == SPR_NORMAL){
			SpriteTypeMap::iterator itr_type = sprite_type_map.find((int)(*i).type);
			if (itr_type != sprite_type_map.end()){
				SpriteImageMap::iterator itr_image = sprite_image_map.find((*itr_type).second);
				if (itr_image != sprite_image_map.end()){
					bImageValid = true;
					bMouseOver = (pdx >= x-128 + (*itr_image).second.rtClip.left - 8)
						&& (pdx < x-128 + (*itr_image).second.rtClip.right + 8)
						&& (pdy >= y-128 + (*itr_image).second.rtClip.top - 8 - 18)
						&& (pdy < y-128 + (*itr_image).second.rtClip.top - 7);
					bMouseOver |= (pdx >= x-128 + (*itr_image).second.rtClip.left - 8)
						&& (pdx < x-128 + (*itr_image).second.rtClip.left - 8 + 5)
						&& (pdy >= y-128 + (*itr_image).second.rtClip.top - 8) && (pdy < y-128 + (*itr_image).second.rtClip.bottom + 8);
					bMouseOver |= (pdx < x-128 + (*itr_image).second.rtClip.right + 8)
						&& (pdx >= x-128 + (*itr_image).second.rtClip.right + 8 - 5)
						&& (pdy >= y-128 + (*itr_image).second.rtClip.top - 8) && (pdy < y-128 + (*itr_image).second.rtClip.bottom + 8);
					bMouseOver |= (pdy >= y-128 + (*itr_image).second.rtClip.top - 8)
						&& (pdy < y-128 + (*itr_image).second.rtClip.top - 8 + 5)
						&& (pdx >= x-128 + (*itr_image).second.rtClip.left - 8) && (pdx < x-128 + (*itr_image).second.rtClip.right + 8);
					bMouseOver |= (pdy < y-128 + (*itr_image).second.rtClip.bottom + 8)
						&& (pdy >= y-128 + (*itr_image).second.rtClip.bottom + 8 - 5)
						&& (pdx >= x-128 + (*itr_image).second.rtClip.left - 8) && (pdx < x-128 + (*itr_image).second.rtClip.right + 8);
				}
			}
		}
		if (!bImageValid){
			bMouseOver = pdx >= x - SPRITE_WIDTH/2 && pdx < x + SPRITE_WIDTH/2 && pdy >= y-SPRITE_HEIGHT && pdy < y;
		}
		if (bMouseOver){
			isel = n;
			SendMessage(GetDlgItem(hWnd, IDC_LIST0), LB_SETCURSEL, (WPARAM)isel, 0);
			sel_changed(n);
			bDragging = true;
			dragging_x = pdx - x;
			dragging_y = pdy - y;
			drag_temp_x = x;
			drag_temp_y = y;
			return true;
		}
	}
	isel = -1;
	return false;
}

void LevelSpritesManager::release_click(bool bApply){
	LevelSpritesData::iterator i = get_iterator(isel);
	if (i == data.end()) return;
	if (bApply) spr_apply(isel);
	bDragging = false;
	return;
}

void LevelSpritesManager::dragging(int pdx, int pdy){
	char s[256];
	LevelSpritesData::iterator i = get_iterator(isel);
	if (i == data.end()) return;
	drag_temp_x = pdx - dragging_x;
	drag_temp_y = pdy - dragging_y;
	sprintf_s(s, sizeof s, "%d", drag_temp_x);
	SetDlgItemText(hWnd, IDC_EDIT2, s);
	sprintf_s(s, sizeof s, "%d", drag_temp_y);
	SetDlgItemText(hWnd, IDC_EDIT3, s);

	sprintf_s(s, sizeof s, "XvCg#%d: (%d, %d)  (%d, %d)", isel, (*i).pdx, (*i).pdy, drag_temp_x, drag_temp_y);
	SendMessage(project->get_hwStatus(), SB_SETTEXT, (WPARAM)STATUS_INDEX_POINTING_OBJECT, (LPARAM)s);
	return;
}

void LevelSpritesManager::release_sel(){
	isel = -1;
	return;
}




BOOL LevelSpritesManager::command(WPARAM wParam, LPARAM lParam){
	if (LOWORD(wParam) == IDM_VIEW_SPRITES){
		switch (iShowSprites){
			case SPR_HIDE:	 iShowSprites = SPR_NORMAL;	break;
			case SPR_NORMAL: iShowSprites = SPR_SIMPLE;	break;
			case SPR_SIMPLE: iShowSprites = SPR_HIDE;	break;
		}
		set_menuitem_state();
		if (iShowSprites == SPR_HIDE){
			isel = -1;
		}
		level->redraw_default();
		return TRUE;
	}
	if (LOWORD(wParam) == IDM_VIEW_SPRITES_I){
		switch (iShowSprites){
			case SPR_HIDE:	 iShowSprites = SPR_SIMPLE;	break;
			case SPR_NORMAL: iShowSprites = SPR_HIDE;	break;
			case SPR_SIMPLE: iShowSprites = SPR_NORMAL;	break;
		}
		set_menuitem_state();
		if (iShowSprites == SPR_HIDE){
			isel = -1;
		}
		level->redraw_default();
		return TRUE;
	}
	if (LOWORD(wParam) == IDM_VIEW_SPR_HIDE){
		iShowSprites = SPR_HIDE;
		set_menuitem_state();
		level->redraw_default();
		return TRUE;
	}
	if (LOWORD(wParam) == IDM_VIEW_SPR_NORMAL){
		iShowSprites = SPR_NORMAL;
		set_menuitem_state();
		level->redraw_default();
		return TRUE;
	}
	if (LOWORD(wParam) == IDM_VIEW_SPR_SIMPLE){
		iShowSprites = SPR_SIMPLE;
		set_menuitem_state();
		level->redraw_default();
		return TRUE;
	}
	return FALSE;
}

void LevelSpritesManager::draw(BitmapBuffer *buf, const RECT *rtRedraw){
	if (iShowSprites == SPR_HIDE) return;
	char s[64];
	int x, y, n;
	POINT points[3];
	LevelSpritesData::iterator i, i_end = data.end();
	buf->select_object(hftSpriteText);
	SetBkMode(buf->get_HDC(), TRANSPARENT);
	for (i=data.begin(), n=0; i!=i_end; i++, n++){
		if (bDragging && n==isel) continue;			// hbO͊YIuWFNgł͕`悵Ȃ
		x = (*i).pdx - rtRedraw->left;
		y = (*i).pdy - rtRedraw->top;

		bool bImageValid = false;
		if (iShowSprites == SPR_NORMAL){
			SpriteTypeMap::iterator itr_type = sprite_type_map.find((int)(*i).type);
			if (itr_type != sprite_type_map.end()){
				SpriteImageMap::iterator itr_image = sprite_image_map.find((*itr_type).second);
				if (itr_image != sprite_image_map.end()){
					bImageValid = true;
					ImageList_Draw(hilSpriteImage, (*itr_image).second.index, buf->get_HDC(), x-128, y-128, ILD_TRANSPARENT);
					buf->select_object(hpnSpriteLine);
					buf->select_object(GetStockObject(NULL_BRUSH));
					buf->rectangle(x-128 + (*itr_image).second.rtClip.left - 8,
						y-128 + (*itr_image).second.rtClip.top - 8 - 18,
						x-128 + (*itr_image).second.rtClip.right + 8,
						y-128 + (*itr_image).second.rtClip.bottom + 8);
					if (n == isel){
						buf->select_object(hpnSpriteSel);
						buf->select_object(GetStockObject(NULL_BRUSH));
						buf->rectangle(x-128 + (*itr_image).second.rtClip.left - 8 + 2,
							y-128 + (*itr_image).second.rtClip.top - 8 + 2,
							x-128 + (*itr_image).second.rtClip.right + 8 - 2,
							y-128 + (*itr_image).second.rtClip.bottom + 8 - 2);
					}
					buf->select_object(hpnSpriteLine);
					buf->select_object(n==isel ? hbrSpriteSel : hbrSpriteRect);
					buf->rectangle(x-128 + (*itr_image).second.rtClip.left - 8,
						y-128 + (*itr_image).second.rtClip.top - 8 - 18,
						x-128 + (*itr_image).second.rtClip.right + 8,
						y-128 + (*itr_image).second.rtClip.top - 7);
					SetTextColor(buf->get_HDC(), n==isel ? RGB(255, 255, 255) : RGB(255, 0, 0));
					sprintf_s(s, sizeof s, "#%d", n);
					TextOut(buf->get_HDC(), x-128+(*itr_image).second.rtClip.left-5,
						y-128 + (*itr_image).second.rtClip.top-23, s, strlen(s));
				}
			}
		}
		if (!bImageValid){
			buf->select_object(hpnSpriteLine);
			buf->select_object(n==isel? hbrSpriteSel : hbrSpriteRect);
			buf->rectangle(x - SPRITE_WIDTH/2, y - SPRITE_HEIGHT, x + SPRITE_WIDTH/2, y);
			points[0].x = x - 6;
			points[0].y = y;
			points[1].x = x;
			points[1].y = y - 6;
			points[2].x = x + 6;
			points[2].y = y;
			buf->select_object(GetStockObject(NULL_PEN));
			buf->select_object(hbrSpriteLine);
			Polygon(buf->get_HDC(), points, 3);
			SetTextColor(buf->get_HDC(), n==isel ? RGB(255, 255, 255) : RGB(255, 0, 0));
			sprintf_s(s, sizeof s, "#%d", n);
			TextOut(buf->get_HDC(), x-SPRITE_WIDTH/2+3, y-SPRITE_HEIGHT+3, s, strlen(s));
			sprintf_s(s, sizeof s, "%04X", (*i).type);
			TextOut(buf->get_HDC(), x-SPRITE_WIDTH/2+3, y-SPRITE_HEIGHT+16, s, strlen(s));
			if (n == isel){
				buf->select_object(hpnSpriteSel);
				buf->select_object(GetStockObject(NULL_BRUSH));
				buf->rectangle(x - SPRITE_WIDTH/2-2, y - SPRITE_HEIGHT-2, x + SPRITE_WIDTH/2+2, y+2);
			}
		}

	}
	return;
}

void LevelSpritesManager::draw_f(BitmapBuffer *buf, const RECT *rtRedraw){
	char s[64];
	int x, y;
	POINT points[3];
	if (iShowSprites == SPR_HIDE) return;
	if (!bDragging) return;
	LevelSpritesData::iterator i = get_iterator(isel);
	if (i == data.end()) return;
	
	SetBkMode(buf->get_HDC(), TRANSPARENT);
	SetTextColor(buf->get_HDC(), RGB(255, 255, 255));
	x = drag_temp_x - rtRedraw->left;
	y = drag_temp_y - rtRedraw->top;
	bool bImageValid = false;
	if (iShowSprites == SPR_NORMAL){
		SpriteTypeMap::iterator itr_type = sprite_type_map.find((int)(*i).type);
		if (itr_type != sprite_type_map.end()){
			SpriteImageMap::iterator itr_image = sprite_image_map.find((*itr_type).second);
			if (itr_image != sprite_image_map.end()){
				bImageValid = true;
				ImageList_Draw(hilSpriteImage, (*itr_image).second.index, buf->get_HDC(), x-128, y-128, ILD_TRANSPARENT);
				buf->select_object(hpnSpriteLine);
				buf->select_object(GetStockObject(NULL_BRUSH));
				buf->rectangle(x-128 + (*itr_image).second.rtClip.left - 8,
					y-128 + (*itr_image).second.rtClip.top - 8 - 18,
					x-128 + (*itr_image).second.rtClip.right + 8,
					y-128 + (*itr_image).second.rtClip.bottom + 8);
				buf->select_object(hpnSpriteSel);
				buf->select_object(GetStockObject(NULL_BRUSH));
				buf->rectangle(x-128 + (*itr_image).second.rtClip.left - 8 + 2,
					y-128 + (*itr_image).second.rtClip.top - 8 + 2,
					x-128 + (*itr_image).second.rtClip.right + 8 - 2,
					y-128 + (*itr_image).second.rtClip.bottom + 8 - 2);
				buf->select_object(hpnSpriteLine);
				buf->select_object(hbrSpriteSel);
				buf->rectangle(x-128 + (*itr_image).second.rtClip.left - 8,
					y-128 + (*itr_image).second.rtClip.top - 8 - 18,
					x-128 + (*itr_image).second.rtClip.right + 8,
					y-128 + (*itr_image).second.rtClip.top - 7);
				sprintf_s(s, sizeof s, "#%d", isel);
				TextOut(buf->get_HDC(), x-128+(*itr_image).second.rtClip.left-5,
					y-128 + (*itr_image).second.rtClip.top-23, s, strlen(s));
			}
		}
	}
	if (!bImageValid){
		buf->select_object(hpnSpriteLine);
		buf->select_object(hbrSpriteSel);
		buf->rectangle(x - SPRITE_WIDTH/2, y - SPRITE_HEIGHT, x + SPRITE_WIDTH/2, y);
		points[0].x = x - 6;
		points[0].y = y;
		points[1].x = x;
		points[1].y = y - 6;
		points[2].x = x + 6;
		points[2].y = y;
		buf->select_object(GetStockObject(NULL_PEN));
		buf->select_object(hbrSpriteLine);
		Polygon(buf->get_HDC(), points, 3);
		sprintf_s(s, sizeof s, "#%d", isel);
		TextOut(buf->get_HDC(), x-SPRITE_WIDTH/2+3, y-SPRITE_HEIGHT+3, s, strlen(s));
		sprintf_s(s, sizeof s, "%04X", (*i).type);
		TextOut(buf->get_HDC(), x-SPRITE_WIDTH/2+3, y-SPRITE_HEIGHT+16, s, strlen(s));
		buf->select_object(hpnSpriteSel);
		buf->select_object(GetStockObject(NULL_BRUSH));
		buf->rectangle(x - SPRITE_WIDTH/2-2, y - SPRITE_HEIGHT-2, x + SPRITE_WIDTH/2+2, y+2);
	}

	return;
}

void LevelSpritesManager::select_sprite(int num){
	char s[128], s2[128];
	isel = num;
	LevelSpritesData::iterator i = get_iterator(num);
	level->select_page(LEVEL_SPRITES);
	level->excluse_pages(LEVEL_SPRITES);
	HWND hwList = GetDlgItem(hWnd, IDC_LIST0);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)FALSE, 0);
	SendMessage(hwList, LB_DELETESTRING, (WPARAM)num, 0);
	get_sprite_text(s, sizeof s, &*i);
	sprintf_s(s2, sizeof s2, "#%d: %s", num, s);
	SendMessage(hwList, LB_INSERTSTRING, (WPARAM)num, (LPARAM)s2);
	SendMessage(hwList, WM_SETREDRAW, (WPARAM)TRUE, 0);
	SendMessage(hwList, LB_SETCURSEL, (WPARAM)num, 0);
	output_sprite_data(&*i, num);
	focus_on_sprite(num);
}

BOOL LevelSpritesManager::translate_dialog_message(MSG *msg){
	return IsDialogMessage(hWnd, msg);
}






