
///////////////////////////////////////////////////////////////////////////////
//
//	fmf_io_mc.cpp
//

#include "platinum_def.h"



bool OutputFMF_MapConv(LPCWSTR filename, const WORD *buf_src, DWORD width, DWORD height, bool bVert){
	HANDLE hFile = CreateFile(filename, GENERIC_WRITE, FILE_SHARE_READ, NULL,
		CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) return false;

	DWORD size_written;
	FMFHeader header;
	header.identifier	= MAKE_IDENTIFIER('F', 'M', 'F', '_');
	header.size			= width * height * 2;
	header.map_width	= width;
	header.map_height	= height;
	header.chip_width	= 32;
	header.chip_height	= 32;
	header.layer_count	= 1;
	header.bit_count	= 16;
	WriteFile(hFile, &header, sizeof header, &size_written, NULL);

	DWORD x, y;
	WORD chip;
	WORD hvflip;

	if (bVert){
		for (y=0; y<height; y++){
			for (x=0; x<width; x++){
				chip = buf_src[x + y*width];
				hvflip = chip & 0xC000;
				chip &= 0x3fff;
				chip = ((chip&0x3ff0)<<4) | (chip&0x000f) | (hvflip >> 10);
				WriteFile(hFile, &chip, 2, &size_written, NULL);
			}
		}
	}else{
		for (y=0; y<height; y++){
			for (x=0; x<width; x++){
				chip = buf_src[x*height + y];
				hvflip = chip & 0xC000;
				chip &= 0x3fff;
				chip = ((chip&0x3ff0)<<4) | (chip&0x000f) | (hvflip >> 10);
				WriteFile(hFile, &chip, 2, &size_written, NULL);
			}
		}
	}

	CloseHandle(hFile);
	return true;
}



bool InputFMF_MapConv(LPCWSTR filename, WORD *buf, DWORD buf_size, DWORD *width, DWORD *height, bool bVert){
	HANDLE hFile = CreateFile(filename, GENERIC_READ, FILE_SHARE_READ, NULL,
		OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if (hFile == INVALID_HANDLE_VALUE) return false;

	DWORD size_read;
	FMFHeader header;
	ReadFile(hFile, &header, sizeof header, &size_read, NULL);
	if (header.identifier != MAKE_IDENTIFIER('F', 'M', 'F', '_')
		|| header.bit_count != 16){
		// `ǂႤ
		CloseHandle(hFile);
		return false;
	}
	*width  = header.map_width;
	*height = header.map_height;
	if ((*width)*(*height)*2 > buf_size){
		// obt@s
		return false;
	}

	DWORD x, y;
	WORD chip;
	WORD hflip;
	WORD vflip;

	if (bVert){
		for (y=0; y<*height; y++){
			for (x=0; x<*width; x++){
				ReadFile(hFile, &chip, 2, &size_read, NULL);
				hflip = chip & 0x0010;
				vflip = chip & 0x0020;
				chip = ((chip & 0xff00) >> 4) | (chip & 0x000f);
				chip |= ((hflip!=0) << 14) | ((vflip!=0) << 15);
				buf[x + y*(*width)] = chip;
			}
		}
	}else{
		for (y=0; y<*height; y++){
			for (x=0; x<*width; x++){
				ReadFile(hFile, &chip, 2, &size_read, NULL);
				hflip = chip & 0x0010;
				vflip = chip & 0x0020;
				chip = ((chip & 0xff00) >> 4) | (chip & 0x000f);
				chip |= ((hflip!=0) << 14) | ((vflip!=0) << 15);
				buf[x*(*height) + y] = chip;
			}
		}
	}

	CloseHandle(hFile);
	return true;
}






